//
//  CwlDispatchTests.swift
//  CwlUtils
//
//  Created by Matt Gallagher on 2016/07/30.
//  Copyright © 2016 Matt Gallagher ( https://www.cocoawithlove.com ). All rights reserved.
//
//  Permission to use, copy, modify, and/or distribute this software for any
//  purpose with or without fee is hereby granted, provided that the above
//  copyright notice and this permission notice appear in all copies.
//
//  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
//  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
//  SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
//  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
//  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
//  IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//

import Foundation
import XCTest
import CwlUtils

class DispatchTests: XCTestCase {
	let queue = DispatchQueue(label: "")
	let value = 3
	let key = DispatchSpecificKey<Int>()
	
	override func setUp() {
		queue.setSpecific(key: key, value: value)
	}
	
	func testSynchronousSingleTimer() {
		do {
			// Ensure falling out of scope cancels the timer
			_ = DispatchSource.singleTimer(interval: .milliseconds(10), queue: queue) {
				XCTFail()
			}
		}
		
		let ex = expectation(description: "")
		var timer: DispatchSourceTimer? = nil
		queue.sync {
			timer = DispatchSource.singleTimer(interval: .milliseconds(50), queue: queue) {
				// Ensure test occurs on the appropriate queue
				XCTAssert(DispatchQueue.getSpecific(key: self.key) != nil)
				ex.fulfill()
			}
		}
		waitForExpectations(timeout: 1e2, handler: nil)
		withExtendedLifetime(timer) {}
	}
	
	func testSynchronousPeriodicTimer() {
		do {
			// Ensure falling out of scope cancels the timer
			_ = DispatchSource.repeatingTimer(interval: .milliseconds(10), queue: queue) {
				XCTFail()
			}
		}
		
		let ex = expectation(description: "")
		var timer: DispatchSourceTimer? = nil
		var fireCount = 0
		queue.sync {
			timer = DispatchSource.repeatingTimer(interval: .milliseconds(50), queue: queue) {
				// Ensure test occurs on the appropriate queue
				XCTAssert(DispatchQueue.getSpecific(key: self.key) == self.value)
				
				fireCount += 1
				if fireCount == 3 {
					ex.fulfill()
				}
			}
		}
		waitForExpectations(timeout: 1e2, handler: nil)
		withExtendedLifetime(timer) {}
	}
	
	func testParametricSingleTimer() {
		do {
			// Ensure falling out of scope cancels the timer
			_ = DispatchSource.singleTimer(parameter: 0, interval: .milliseconds(10)) { p in
				XCTFail()
			}
		}
		
		let ex1 = expectation(description: "")
		let ex2 = expectation(description: "")
		var timer: DispatchSourceTimer? = nil
		var outerMutexComplete = false
		
		// Test rescheduling a timer during the callback for an earlier scheduling
		queue.sync {
			timer = DispatchSource.singleTimer(parameter: 1, interval: .milliseconds(1)) { p in
				XCTAssert(!outerMutexComplete)
				self.queue.sync {
					XCTAssert(outerMutexComplete)
					XCTAssert(p == 1)
					ex1.fulfill()
				}
			}
			Thread.sleep(forTimeInterval: 0.1)
			timer?.scheduleOneshot(parameter: 2, interval: .milliseconds(1)) { p in
				XCTAssert(outerMutexComplete)
				self.queue.sync {
					XCTAssert(outerMutexComplete)
					XCTAssert(p == 2)
					ex2.fulfill()
				}
			}
			Thread.sleep(forTimeInterval: 0.1)
			outerMutexComplete = true
		}
		waitForExpectations(timeout: 1e2, handler: nil)
		withExtendedLifetime(timer) {}
	}
	
	func testParametricPeriodicTimer() {
		do {
			// Ensure falling out of scope cancels the timer
			_ = DispatchSource.repeatingTimer(parameter: 0, interval: .milliseconds(10)) { p in
				XCTFail()
			}
		}
		
		let ex1 = expectation(description: "")
		let ex2 = expectation(description: "")
		var timer: DispatchSourceTimer? = nil
		var outerMutexComplete = false
		var fireCount = 0
		
		// Test rescheduling a timer during the callback for an earlier scheduling
		queue.sync {
			timer = DispatchSource.repeatingTimer(parameter: 1, interval: .milliseconds(1)) { p in
				XCTAssert(!outerMutexComplete)
				self.queue.sync {
					XCTAssert(outerMutexComplete)
					XCTAssert(p == 1)
					ex1.fulfill()
				}
			}
			Thread.sleep(forTimeInterval: 0.1)
			timer?.scheduleRepeating(parameter: 2, interval: .milliseconds(1)) { p in
				XCTAssert(outerMutexComplete)
				self.queue.sync {
					XCTAssert(outerMutexComplete)
					XCTAssert(p == 2)
					
					fireCount += 1
					if fireCount == 3 {
						ex2.fulfill()
					}
				}
			}
			Thread.sleep(forTimeInterval: 0.1)
			outerMutexComplete = true
		}
		waitForExpectations(timeout: 1e2, handler: nil)
		withExtendedLifetime(timer) {}
	}
	
	func testFromTimeInterval() {
		let d = DispatchTimeInterval.interval(1.125)
		XCTAssert(d.seconds == 1.125)
	}
	
	func testToSeconds() {
		let s = DispatchTimeInterval.seconds(1).seconds
		XCTAssert(s > 0.99999999 && s < 1.00000001)
		
		let m = DispatchTimeInterval.milliseconds(1).seconds
		XCTAssert(m > 0.99999999e-3 && m < 1.00000001e-3)
		
		let u = DispatchTimeInterval.microseconds(1).seconds
		XCTAssert(u > 0.99999999e-6 && u < 1.00000001e-6)
		
		let n = DispatchTimeInterval.nanoseconds(1).seconds
		XCTAssert(n > 0.99999999e-9 && n < 1.00000001e-9)
	}
	
	func testToNanoseconds() {
		let s = DispatchTimeInterval.seconds(1).nanoseconds
		XCTAssert(s == Int64(1e9))
		
		let m = DispatchTimeInterval.milliseconds(1).nanoseconds
		XCTAssert(m == Int64(1e6))
		
		let u = DispatchTimeInterval.microseconds(1).nanoseconds
		XCTAssert(u == Int64(1e3))
		
		let n = DispatchTimeInterval.nanoseconds(1).nanoseconds
		XCTAssert(n == Int64(1))
	}
}
