//
//  CwlUnanticipatedErrorTests.swift
//  CwlUtils
//
//  Created by Matt Gallagher on 2015/03/05.
//  Copyright © 2015 Matt Gallagher ( https://www.cocoawithlove.com ). All rights reserved.
//
//  Permission to use, copy, modify, and/or distribute this software for any
//  purpose with or without fee is hereby granted, provided that the above
//  copyright notice and this permission notice appear in all copies.
//
//  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
//  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
//  SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
//  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
//  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
//  IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//

import Foundation
import XCTest
import CwlUtils

enum TestCode: Int, Error {
	case zeroValue = 0
	case oneValue = 1
	case testValue = 2
}

#if os(iOS) || os(macOS)

class UnanticipatedErrorTests: XCTestCase {
	func testUnanticipatedError() {
		let e = TestCode.testValue.withUnanticipatedErrorRecoveryAttempter()
		
		// NOTE: error codes equal raw values only if the raw values are zero based with no gaps
		XCTAssert(e.code == TestCode.testValue.rawValue)
		
		let userInfo = e.userInfo
		if let callStackSymbols = userInfo[UnanticipatedErrorRecoveryAttempter.callStackSymbols] as? [String] {
			XCTAssert(callStackSymbols.count > 1, "No call stack symbols present")
		} else {
			XCTFail("Call stack symbols not present")
		}
		
		XCTAssertNotNil(userInfo[NSLocalizedRecoverySuggestionErrorKey])
		XCTAssertNotNil(userInfo[NSLocalizedRecoveryOptionsErrorKey])
		XCTAssertNotNil(userInfo[NSRecoveryAttempterErrorKey])
		
		XCTAssert((userInfo[NSLocalizedRecoveryOptionsErrorKey] as? [Any])?.count == 2, "Wrong number of options")
		
		let attempter = userInfo[NSRecoveryAttempterErrorKey] as? NSObject
		let backup = pasteboardBackup()
		attempter?.attemptRecovery(fromError: e as NSError, optionIndex: 1)
		let clipboardString = pasteboardString()
		
		// The following (ugly) compile-time conditional is a best effort at testing for the simulator (no actual simulator macro is provided)
		XCTAssert(clipboardString?.range(of: e.localizedRecoverySuggestion!) != nil, "Simulator pasteboard will fail prior to iOS 10/Xcode 8")
		
		restorePasteboard(backup)
	}
}

#endif

#if os(iOS)
	
	func pasteboardBackup() -> [Dictionary<String, NSObject>] {
		return ((UIPasteboard.general.items as NSArray).copy() as? [Dictionary<String, NSObject>]) ?? Array<Dictionary<String, NSObject>>()
	}
	
	func restorePasteboard(_ items: [Dictionary<String, NSObject>]) {
		UIPasteboard.general.items = items
	}
	
	func pasteboardString() -> String? {
		return UIPasteboard.general.string
	}
	
#elseif os(macOS)
	
	func pasteboardBackup() -> [NSPasteboardItem] {
		#if swift(>=4)
			let general = NSPasteboard.general
		#else
			let general = NSPasteboard.general()
		#endif
		return general.pasteboardItems?.map { item in
			let backupItem = NSPasteboardItem()
			for type in item.types {
				if let data = (item.data(forType: type) as NSData?)?.copy() as? Data {
					backupItem.setData(data, forType: type)
				}
			}
			return backupItem
		} ?? []
	}
	
	func restorePasteboard(_ items: [NSPasteboardItem]) {
		#if swift(>=4)
			let general = NSPasteboard.general
		#else
			let general = NSPasteboard.general()
		#endif
		general.clearContents()
		general.writeObjects(items)
	}
	
	func pasteboardString() -> String? {
		#if swift(>=4)
			return NSPasteboard.general.string(forType: NSPasteboard.PasteboardType.string)
		#else
			return NSPasteboard.general().string(forType: NSPasteboardTypeString)
		#endif
	}
	
#endif
